// GSMSMS Library
// Copyright (c) 2014, 2015, 2016, 2017, 2018, 2019 by Haute Solutions / Ed Van Every (www.haute-solutions.com)
//
// FUNCTION: This Library is intended to provide GSM SMS and GPS functionality
//
// HARDWARE: This LIbrary Supports the following GSM chipsets
//						
//					 SIMCOM SIM808
//					 SIMCOM SIM900
//					 SIMCOM SIM5320
//					 SIMCOM SIM7100
//					 SIMCOM SIM7600
//
// As Implemented on the following Boards:
//
//           Itead GBoard (SIM900 on Pins 2/33, Power on Pin 6)
//           Haute Solutions GBoardHS (SIM900 on UART2, Power on Pin 6)
//           Haute Solutions EGGBoard (SIM900 on UART2, Power on Pin 6)
//           Haute Solutions EGGBoard 3G (SIM5320 on UART1, Power on Pin 6)
//           Haute Solutions MeGGaBoard (SIM808 on UART1, Power on Pin 46)
//           Haute Solutions MeGGaBoard 3G (SIM5320 on UART1, Power on Pin 46)
//           Haute Solutions ESP324G (SIM7600 on UART1, Power on Pin 4)
//           Haute Solutions ESP324GE (SIM7600 on UART1, Power on Pin 2)
//
// SOFTWARE: None Additional
//
// LICENSE:  This library is free software: you can redistribute it and/or modify
//           it under the terms of the GNU General Public License as published by
//           the Free Software Foundation, either version 3 of the License, or
//           (at your option) any later version.
//
//           This program is distributed in the hope that it will be useful,
//           but WITHOUT ANY WARRANTY; without even the implied warranty of
//           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//           GNU General Public License for more details.
//           <http://www.gnu.org/licenses/>.
//

#define GSMSMS_LIBVER "4.3"

#include "Arduino.h"
#include "GSMSMS.h"

// Software Serial Dependencies
#ifdef GSM_SW_SERIAL
  #include <SoftwareSerial.h>
  #define GSM_TXPIN 2															// TX Pin used by GBoard with SoftwareSerial
  #define GSM_RXPIN 3															// RX Pin used by GBoard with SoftwareSerial
  SoftwareSerial GSMSS(GSM_TXPIN, GSM_RXPIN);
#endif 

GSMSMS::GSMSMS(int OnPin, int UART, int OnPulse, int OnWait)
{
//  This constructor can be called by specifically designating the Power Control Pin and the UART
//  or it can use the built in board definitions (OnPin >= 9000)

  switch (OnPin) {
  	case GBOARD:
  		_Board = OnPin;
  		_OnPin = 6;
			_UART = 0;
			_OnPulse = 4000;
			_OnWait = 13000;
  		break;
  	case GBOARDHS:
  		_Board = OnPin;
  		_OnPin = 6;
			_UART = 2;
			_OnPulse = 4000;
			_OnWait = 13000;
  		break;
  	case EGGBOARD:
  		_Board = OnPin;
  		_OnPin = 6;
			_UART = 2;
			_OnPulse = 4000;
			_OnWait = 13000;
  		break;
  	case EGGBOARD3G:
    	_Board = OnPin;
			_OnPin = 6;
			_UART = 1;
			_OnPulse = 4000;
			_OnWait = 13000;
  		break;
  	case MEGGABOARD:
    	_Board = OnPin;
			_OnPin = 46;
			_UART = 1;
			_OnPulse = 4000;
			_OnWait = 13000;
  		break;
  	case MEGGABOARD3G:
  		_Board = OnPin;
  		_OnPin = 46;
			_UART = 1;
			_OnPulse = 180;
			_OnWait = 13000;
  		break;
  	case ESP324G:
    	_Board = OnPin;
			_OnPin = 4;
			_UART = 1;
			_RXPin = 18;
			_TXPin = 19;
			_OnPulse = 400;
			_OnWait = 13000;
  		break;
  	case ESP324GE:
    	_Board = OnPin;
			_OnPin = 2;
			_UART = 1;
			_RXPin = 14;
			_TXPin = 4;
			_OnPulse = 400;
			_OnWait = 13000;
  		break;
  	default:
  		_Board = 0;
  		_OnPin = OnPin;
			_UART = UART;
			_OnPulse = OnPulse;
			_OnWait = OnWait;
			break;
	}
	
	pinMode(_OnPin, OUTPUT);
	
	_LastBaud = 0;

}

int GSMSMS::SendCmd(const char *AT_cmd_string)
{
  return SendCmdParams(AT_cmd_string, 1000, 1000, "OK", 10);
}

int GSMSMS::SendCmdParams(const char *AT_cmd_string, int start_comm_tmout, int max_interchar_tmout, const char *response_string, int no_of_attempts) 
{  
int i;
int stat;

  for (i = 0; i < no_of_attempts; i++) {
    if (i > 0) delay(500); 

    Println(AT_cmd_string);
    stat = WaitResp(start_comm_tmout, max_interchar_tmout, response_string);
    if(stat < 2) return stat;
  }
  
  return 2;
}

int GSMSMS::WaitResp(unsigned long start_comm_tmout, unsigned long max_interchar_tmout, const char *response_string)
{
int bytes_rcvd;
int bytes_avail;
unsigned long char_timer;
unsigned long time_elapsed;

    char_timer = millis();
    GSMCommBuff[0] = 0x00;
    bytes_rcvd = 0;
    do {
    	time_elapsed = millis() - char_timer;
      bytes_avail = Available();
      if(bytes_avail == 0) {
      	
    	  if(bytes_rcvd == 0) {
    	  	if( time_elapsed > start_comm_tmout ) break;
        } else {
        	if( time_elapsed > max_interchar_tmout ) break;
        }
        
      } else {

        char_timer = millis();
      
        while(bytes_avail) {
          bytes_avail--;
          if(bytes_rcvd < (GSMBUFFSIZE-1)) {
            GSMCommBuff[bytes_rcvd] = Read();
            GSMCommBuff[++bytes_rcvd] = '\0'; 
          } else Read();      
        }
        
      }      
    } while(1);

    if(bytes_rcvd > 0) {
      if(strstr(GSMCommBuff,response_string) != NULL) return 0;		// Got What we were looking for!
      else return 1;																							// Got Something Else....
    }
    
    return 2;    																									// Got Nothing or Timeout!
}

int GSMSMS::Init(long baud_rate, int debug)
{
// Return Codes
// 0 = OK
// 1 = Unable to set Baud Rate
// 2 = Unknown Chipset
// 3 = General Failure

int i;
int mstat;
long baud;
char *p1;

	_Debug = debug;

	if(_Debug) {
		Serial.print("Requested Baud Rate: ");
		Serial.println(baud_rate);
	}
	Begin(baud_rate);
		
	// Try "Hot Start"
	mstat = CheckModem();
	if(mstat == 2) {																				// Got an "AT" Back, so GSM is up and Baud is good!
		if(_Debug) Serial.println("Hot Start: SUCCEED");
  	if(GetChipset() == 0) return 2;
		return 0;		
	}
	
	if(mstat == 1) {																				// Got an "SOMETHING" Back, so GSM is up but BAUD may not be correct!
		if(_Debug) Serial.println("Hot Start: PARTIAL");
	}

	if(mstat == 0) {																				// Got Nothing Back, so chipset may be "OFF"
		if(_Debug) Serial.println("Hot Start: FAIL");
//		for(i=0; i<3; i++){
				if(_Debug) Serial.println("Turning ON");
				TurnOn(_OnPulse, _OnWait);							         	// Try Turning on
				WaitResp(1000, 500, "Anything");				          // Eat any junk in the comm buffer
//			mstat = CheckModem();															// Are we on?
//			if(mstat > 0) break;
//		}
	}
	
	mstat = CheckModem();
	if(mstat == 2) {																				// Got an "AT" Back, so GSM is up and Baud is good!
		if(_Debug) Serial.println("Cold Start: SUCCEED");
  	if(GetChipset() == 0) return 2;
		return 0;		
	}

	if(_Debug) Serial.println("Baud: Matching");		
	for(i=0; i<4; i++) {

		baud = DetectBaud();
		if(baud == 0) continue;																// Not up yet!  (Retry!)
			
		if(_Debug)  {
			Serial.print("Baud: ");
			Serial.println(baud);
		}
			
		if(baud != baud_rate) {
			if(SetBaud(baud_rate)) return 1;										// Set desired Baud rate
		}
		
		if(GetChipset() == 0) return 2;
		
		return 0;
	}

  return 3;
}

int GSMSMS::Init1(long baud_rate, int debug)
{
// Return Codes
// 0 = OK
// 1 = Unable to set Baud Rate
// 2 = Unknown Chipset
// 3 = General Failure

int i;
long baud;
char *p1;

	_Debug = debug;

	if(_Debug) {
		Serial.print("Baud Rate: ");
		Serial.println(baud_rate);
	}
	Begin(baud_rate);
	
	// Try "Hot Start"
  SendCmd("AT");
	p1=strstr(GSMCommBuff,"OK");
  if(p1 != NULL) {  	
		if(_Debug) Serial.println("Hot Start: SUCCEED");
  	if(GetChipset() != 0) return 0;
  }
	if(_Debug) Serial.println("Hot Start: FAIL");

	TurnOn(_OnPulse, _OnWait);							// Try Turning on
	WaitResp(1000, 500, "Anything");				// Eat any junk in the comm buffer
		
  if(SendCmdParams("AT", 500, 500, "OK", 3) == 0) {
		if(_Debug) Serial.println("Baud Matched");
  	if(GetChipset() != 0) return 0;
  }
	if(_Debug) Serial.println("Baud Not Matched");
		
	for(i=0; i<4; i++) {

		baud = DetectBaud();
		if(baud == 0) continue;								// Not up yet!  (Retry!)
			
		if(_Debug)  {
			Serial.print("Baud: ");
			Serial.println(baud);
		}
			
		if(baud != baud_rate) {
			if(SetBaud(baud_rate)) return 1;		// Set desired Baud rate
		}
		
		if(GetChipset() == 0) return 2;
		
		return 0;
	}

  return 3;
}

int GSMSMS::Init2(long baud_rate, int debug)
{
// Return Codes
// 0 = OK
// 1 = Unable to set Baud Rate
// 2 = Unknown Chipset
// 3 = General Failure

int i;
int ms_pulse;													   		// Turn On Pulse (TBD)
int ms_wait;											 					// Then Wait
long baud;

	_Debug = debug;

	if(_Debug) {
		Serial.print("Requested Baud: ");
		Serial.println(baud_rate);
	}
		
	// Try "Hot Start"
  Begin(baud_rate);
  if(SendCmdParams("AT", 500, 500, "OK", 3) == 0) {
		if(_Debug) Serial.println("Hot Start: SUCCEED");
  	if(GetChipset() != 0) return 0;
  }
	if(_Debug) Serial.println("Hot Start: FAIL");

	for(i=0; i<4; i++) {
		
		if(_Debug) {
			Serial.print("TurnOn Pulse: ");
			Serial.println(ms_pulse);
		}
		
		TurnOn(_OnPulse, _OnWait);						// Try Turning on
		WaitResp(1000, 500, "Anything");			// Eat any junk in the comm buffer
			
		if(_Debug) Serial.println("Detecting Baud...");

		baud = DetectBaud();
		if(baud == 0) continue;								// Not up yet!  (Retry!)
			
		if(_Debug)  {
			Serial.print("Baud: ");
			Serial.println(baud);
		}
			
		if(baud != baud_rate) {
			if(SetBaud(baud_rate)) return 1;		// Set desired Baud rate
		}
		
		if(GetChipset() == 0) return 2;
			
		return 0;
  }
 
  return 3;
}

int GSMSMS::Init3(long baud_rate)
{
int i;
int cmdstat;
long bauds[8]={1200, 2400, 4800, 9600, 19200, 38400, 57600, 115200};
 
  pinMode(_OnPin, OUTPUT);

  Begin(9600);
  
  // Try three times to turn on Module	
  for(i=0; i<3; i++){
    cmdstat=SendCmd("AT");  // 0="OK", 1=Something Else (Bad Baud?), 2=No Response
    if(cmdstat == 2) {
			digitalWrite(_OnPin, HIGH);
			delay(1200);
			digitalWrite(_OnPin, LOW);
			delay(10000);
    } else break;
  }
  
  // If its responding at 9600 and thats our target baud rate we are done!	
  if(SendCmd("AT") == 0){
    if(baud_rate = 9600) return 0; 
  }
 
  // Roll through the potential baud rates and try to reset to 9600 so we can talk with it
  for(i=0; i<8; i++){
    Begin(bauds[i]);
    Println("AT+IPR=9600");
  }
  
  // Check to see if its NOW responding at 9600
  Begin(9600);
  if(SendCmd("AT") > 0) return 1;  // Won't Reply
  
  // If thats our target baud we are done!
  if(baud_rate = 9600) return 0;
 
  // Set Module to desired baud rate
  Print("AT+IPR=");
  Print(baud_rate);
  Print("\r");
  delay(500);
  
  // Check if we got it! 
  Begin(baud_rate);
  delay(100);
  if(SendCmd("AT") == 0) return 0;

  // Otherwise we have failed...
  return 2;
}

void GSMSMS::TurnOn(int ms_high, int ms_low)
{
//Serial.println("Turning On");

	digitalWrite(_OnPin, HIGH);
	delay(ms_high);
	digitalWrite(_OnPin, LOW);
	delay(ms_low);
}

void GSMSMS::TurnOff()
{

	switch (_SimCom) {
		case 808:
		case 900:
  		SendCmd("AT+CPOWD=1"); 
			break;
						
		case 5320:
		case 7100:
		case 7600:
  		SendCmd("AT+CPOF"); 
			break;
	}
  digitalWrite(_OnPin, HIGH);
}	

int GSMSMS::CheckModem() {
int i;
int mstat;
char *p1;

	for(i=0; i<3; i++){
		mstat = SendCmdParams("AT", 500, 500, "OK", 3);
		switch (mstat) {
			case 0:								// 0 = Got "OK" (Baud Correct)) - Good to go!
				return 2;
			case 1:								// 1 = Got "Something" (Bad Baud?) - try again (AutoBaud)
				break;
			case 2:								// 2 = No Response (Off)
				return 0;
		}
	}
	
	return mstat;	
}

long GSMSMS::DetectBaud()
{
int i;
int cmdstat;
long bauds[8]={115200, 9600, 1200, 2400, 4800, 19200, 38400, 57600};

  for(i=0; i<8; i++){
  	if(_Debug)  {
			Serial.print("Trying Baud: ");
			Serial.println(bauds[i]);
		}
    Begin(bauds[i]);
    if(SendCmdParams("AT", 500, 500, "OK", 3) == 0) return bauds[i];
  }
  
	return 0;
}

int GSMSMS::SetBaud(long baud_rate)
{
  Print("AT+IPR=");
  Print(baud_rate);
  Print("\r");
  delay(5000);
  
  // Check if we got it! 
  Begin(baud_rate);
  delay(100);
  if(SendCmd("AT") == 0) return 0;

  // Otherwise we have failed...
  return 1;
}

void GSMSMS::SetParams() 
{
  SendCmdParams("ATE0", 500, 50, "OK", 5);                                   	// switch off echo
  SendCmdParams("AT+CLIP=1", 500, 50, "OK", 5);                              	// Request calling line identification
  SendCmdParams("AT+CMEE=0", 500, 50, "OK", 5);		                						// Mobile Equipment Error Code
  SendCmdParams("AT+CMGF=1", 500, 50, "OK", 5);		                						// set the SMS mode to text 
  SendCmdParams("AT+CPBS=\"SM\"", 1000, 50, "OK", 5);                       	// select phonebook memory storage
  
  SendCmdParams("AT+CNMI=0", 1000, 50, "OK", 2);                             	// Disable messages about new SMS from the GSM module 
  SendCmdParams("AT+CPMS=\"SM\",\"SM\",\"SM\"", 1000, 1000, "+CPMS:", 10);  	// Init memory for SMS in the SIM card
  
  switch (_SimCom) {
  	case 808:
  	case 900:
  		SendCmdParams("AT+CLTS=1", 500, 50, "OK", 5);                          	// Use Network Time
  		SendCmdParams("AT+CSDT=1", 10000, 1000, "OK", 5);                      	// Enable SIM Card Detection
  		break;
  	case 5320:
  	case 7100:
  	case 7600:
  		SendCmdParams("AT+CTZU=1", 500, 50, "OK", 5);                          	// Use Network Time
  		break;
  }
  
//SendCmdParams("AT&F", 1000, 50, "OK", 5);                            				// Reset to the factory settings    
//SendCmdParams("AT+CIPSHUT", 500, 50, "SHUT OK", 5);
//SendCmdParams("AT+IPR=9600", 500, 50, "OK", 5);		                					// setup fixed baud rate
//SendCmdParams("AT#SELINT=1", 500, 50, "OK", 5);		                					// setup mode
//SendCmdParams("AT#GPIO=8,1,1", 500, 50, "OK", 5);		                				// Switch ON User LED - just as signalization we are here
//SendCmdParams("AT#GPIO=9,0,0", 500, 50, "OK", 5);		                				// Sets GPIO9 as an input = user button
//SendCmdParams("AT#GPIO=5,0,2", 500, 50, "OK", 5);		                				// allow audio amplifier control
//SendCmdParams("AT#GPIO=8,0,1", 500, 50, "OK", 5);		                				// Switch OFF User LED- just as signalization we are finished
//SendCmdParams("AT#SHFEC=1", 500, 50, "OK", 5);		                					// Echo canceller enabled 
//SendCmdParams("AT#SRS=26,0", 500, 50, "OK", 5);		                					// Ringer tone select (0 to 32)
//SendCmdParams("AT#HFMICG=7", 1000, 50, "OK", 5);		                				// Microphone gain (0 to 7) - response here sometimes takes more than 500msec. so 1000msec. is more safety		
//SendCmdParams("ATS0=1", 500, 50, "OK", 5);		                        			// Auto answer after first ring enabled - auto answer is not used
//SendCmdParams("AT#SRP=1", 500, 50, "OK", 5);		                						// select ringer path to handsfree
//SendCmdParams("AT+CRSL=2", 500, 50, "OK", 5);		                						// select ringer sound level  
}

int GSMSMS::GetSMS(char *number, char *msg, int nsize, int msize) 
{
char *p1;
char *p2;
int msgno;
int i;
  
//if(SendCmd("AT+CMGL=\"REC UNREAD\"") > 0) return 0;
  SendCmd("AT+CMGL=\"ALL\"");
  
  p1=strstr(GSMCommBuff,"+CMGL:");
  if(p1 == NULL) return 0;
  msgno=atoi(p1+6); 

  Print("AT+CMGR=");
  Println(msgno);
  if(WaitResp(1000, 1000, "OK") > 0) return 0;
  
  // CALLER will be between THIRD Parenthesis and FOURTH Parenthesis
  p1=strstr(GSMCommBuff,"+CMGR:");
  for(i=0; i<3; i++) {
    p1 = strchr(p1,0x22);
    if(p1 == NULL) return 0;
    p1++;
  }
  for(i=0; i<nsize; i++) {
  	if(p1[i] != 0x22) number[i]=p1[i];
  	else {
  		number[i]='\0';
  		break;
  	}
  }
  number[nsize-1]='\0';
  p1++;
   
	// MESSAGE will start on the second line
  p1=strchr(p1,0x0A);
  if(p1 != NULL) {
    p1++;
    
    strncpy(msg,p1,msize);
    msg[msize-1]='\0';
    
    // Trim trailing CR/LF/"OK"...
    i=strlen(msg);
    if(!strcmp(msg+(i-8),"\r\n\r\nOK\r\n")) msg[i-8]='\0';
    	
    // Trim any remaining trailing CR's or LF's...
    for(i=strlen(msg)-1; i >= 0; i--) {
    	if((msg[i] == 0x0A) || (msg[i] == 0x0D)) msg[i]='\0';	
     	else break;
		}
        	
  } else return 0;

  Print("AT+CMGD=");
  Println(msgno);
  WaitResp(1000, 1000, "OK");
  
  return 1;			// This used to return the MSGNO of the message being received except that the SIM5320 starts numbering at "zero"!!!!
}

int GSMSMS::SendSMS(const char *number, const char *msg)
{
char msgend[2]={0x1a,'\0'};

  if(strlen(number) > GSMBUFFSIZE - 11) return 1;
  if(strlen(msg) > GSMBUFFSIZE - 2) return 2;
  
  if(PrepToSendSMS(number) > 0) return 3;
  Print(msg);
  Println(msgend);
  if(FlushSMS() > 0) return 4;  
  return 0;
}

int GSMSMS::PrepToSendSMS(const char *number)
{
  Print("AT+CMGS=\"");
  Print(number);
  Println("\"");
  
  return WaitResp(1000,500,">");
}

int GSMSMS::FlushSMS()
{
    return WaitResp(7000,5000,"+CMGS");
}

void GSMSMS::DeleteAllSMS() {
	switch (_SimCom) {
		case 808:
		case 900:
  		SendCmd("AT+CMGDA=\"DEL ALL\""); 
			break;
						
		case 5320:
		case 7100:
		case 7600:
  		SendCmd("AT+CMGD=0,4"); 
			break;
	}
}

int GSMSMS::GetSignalQuality() {
char *p1;
char *p2;
int csq;

  SendCmd("AT+CSQ");
  
  // Isolate just the Return Code in the Comm Buffer
  p1=strstr((char *)GSMCommBuff,"+CSQ:");
  if(p1 == NULL) return 0;
  p2=strchr(p1,'\n');
  if(p2 == NULL) return 0;
  *p2='\0';
  
  // Retrieve SIGNAL QUALITY starting at 7th char and ending before the Comma
  p1=p1+6;
  p2=strchr(p1,',');
  if(p2 == NULL) return 0;
  *p2='\0';
  
  csq=atoi(p1);
  
  if(csq == 99) return 0;
  
  if(csq > 19) return 5;
  if(csq > 14) return 4;
  if(csq > 9)  return 3;
  if(csq > 5)  return 2;
  if(csq > 1)  return 1;
  
  return 0;	
}

int GSMSMS::GetTime(char *time) {
char *p1;
char *p2;
char t[3];
int year;

  SendCmd("AT+CCLK?");
  
  // Isolate just the Return Code in the Comm Buffer
  p1=strstr((char *)GSMCommBuff,"+CCLK:");
  if(p1 == NULL) return 0;
  p2=strchr(p1,'\n');
  if(p2 == NULL) return 0;
  *p2='\0';
  
  // Retrieve TIME between FIRST Parenthesis and SECOND COLON (Trim seconds and TimeZone)
  p1=strchr(p1,0x22);
  if(p1 == NULL) return 0;
  p1++;
  p2=strchr(p1,':');
  if(p2 == NULL) return 0;
  p2=strchr(p2+1,':');
  if(p2 == NULL) return 0;
  *p2='\0';
  
  strcpy(time,p1);
  
  // Check if the time is valid (if not, SMS Message should be sent to phone)
  strncpy(t,time,2);
  year=atoi(t);
  if(year < 12) return 0;  
  
  return 1;
}

int GSMSMS::SIMInserted() {
char *p1;
char *p2;

	switch (_SimCom) {
		case 808:
		case 900:
  		SendCmd("AT+CSMINS?");
  
  		// Isolate just the Return Code in the Comm Buffer
  		p1=strstr(GSMCommBuff,"+CSMINS:");
  		if(p1 == NULL) return 0;
  		p2=strchr(p1,'\n');
  		if(p2 == NULL) return 0;
  		*p2='\0';
  
  		// Insertion Status is the character after the first comma (0 = not inserted, 1 = inserted)
  		p1=strchr(p1,',');
  		if(p1 == NULL) return 0;
  		p1++;

			return (*p1 - 0x30);
			
		case 5320:
  	case 7100:
		case 7600:
			SendCmd("AT+CICCID");			// Request ICCID
  
  		// Did we get an ERROR?
  		p1=strstr(GSMCommBuff,"ERROR");
  		if(p1 == NULL) return 1;
	}
	
	return 0;
}

int GSMSMS::RegisteredOnNetwork() {
int RegStat3G;
int RegStat4G;

	_GService = 0;
	
	if(_SimCom < 7000) {																				//  If this is a 3G chipset, simply return the 3G registration status
		RegStat3G = RegisteredOnNetworkExt(3);
		if(RegStat3G == 1 || RegStat3G == 5) _GService=3;
		return RegStat3G;
	}
	
	RegStat4G=RegisteredOnNetworkExt(4);												//  Must be a 4G chipset so get 4G status
	if(RegStat4G == 1 || RegStat4G == 5) {											//  If Registered 4G we are good to go!
		_GService = 4;
		return RegStat4G;
	}

	RegStat3G=RegisteredOnNetworkExt(3);												//  Otherwise, check to see if we are registered on 3G network...
		
	if(RegStat3G == 1 || RegStat3G == 5) {											//  If registered under 3G then return 3G code  (as 4G status may be 4/Unknown)
		_GService = 3;
		return RegStat3G;
	}
	
	return RegStat4G;																						//  Otherwise return 4G status
}

int GSMSMS::RegisteredOnNetworkExt(int mode) {
char *p1;
char *p2;

	switch (mode) {
		case 4:				// 4G
			SendCmd("AT+CEREG?");
  		p1=strstr(GSMCommBuff,"+CEREG:");			// Isolate just the Return Code in the Comm Buffer
			break;
		default:			// 2G or 3G
			SendCmd("AT+CREG?");
  		p1=strstr(GSMCommBuff,"+CREG:");			// Isolate just the Return Code in the Comm Buffer
	}
  if(p1 == NULL) return 0;
  p2=strchr(p1,'\n');
  if(p2 == NULL) return 0;
  *p2='\0';
  
  // Registration Status is the character after the first comma
  // 0 = Not Registered
  // 1 = Registered on Home Network
  // 2 = Not Registered, still searching
  // 3 = Registration Denied
  // 4 = Unknown
  // 5 = Registered, Roaming
  // 6 = Registered for "SMS only", home network (4G)
  // 7 = Registered for "SMS only", roaming (4G)
  // 8 = Attached for emergency bearer services only (4G)
  
  p1=strchr(p1,',');
  if(p1 == NULL) return 0;
  p1++;

	return (*p1 - 0x30);
}

int GSMSMS::GetGService() {
	return _GService;			// 3G=3, 4G=4
}

int GSMSMS::GetSubscriberNumber(char *snum) {
char *p1;
char *p2;

	strcpy(snum,"");

  SendCmd("AT+CNUM");
  
  // Isolate just the Return Code in the Comm Buffer
  p1=strstr(GSMCommBuff,"+CNUM:");
  if(p1 == NULL) return 0;
  p2=strchr(p1,'\n');
  if(p2 == NULL) return 0;
  *p2='\0';
  
  // Retrieve the Subscriber Number after first Comma and between Parenthesis
  p1=strchr(p1,',');
  if(p1 == NULL) return 0;
  p1=strchr(p1,'\"');
  if(p1 == NULL) return 0;
  p1++;
  p2=strchr(p1,'\"');
  if(p2 == NULL) return 0;
  *p2='\0';
  strcpy(snum,p1);
  
  return 1;
 }

int GSMSMS::GetSimCom() {
	return _SimCom;
}

int GSMSMS::GetChipset(char *chipinfo) {
char *p1;
char *p2;
int chipcode;

  SendCmd("AT+CGMM");
  
  // Isolate just the Return Code in the Comm Buffer
  p1=strstr(GSMCommBuff,"SIMCOM_");
  if(p1 == NULL) return 0;
  p1+=7;
  
  p2=strchr(p1,'\r');
  if(p2 != NULL) *p2='\0';
  p2=strchr(p1,'\n');
  if(p2 != NULL) *p2='\0';

  if(chipinfo != NULL) strcpy(chipinfo,p1);
  	
  if(strncmp(p1,"SIM",3)) _SimCom=0;
  else _SimCom =atoi(p1+3);
  
  return  _SimCom;
	
}

int GSMSMS::GetIMEI(char *imei) {

  switch (_SimCom) {
		case 808:
		case 900:
  		GetATResult("AT+GSN",NULL,imei,1000);
			break;
						
		case 5320:
		case 7100:
		case 7600:
  		GetATResult("AT+SIMEI?","+SIMEI: ",imei,1000);
			break;
	}


  return 1;
}

int GSMSMS::GetICCID(char *iccid) {

	switch (_SimCom) {
		case 808:
		case 900:
  		GetATResult("AT+CCID",NULL,iccid,1000);
			break;
						
		case 5320:
		case 7100:
		case 7600:
  		GetATResult("AT+CICCID","+ICCID: ",iccid,1000);
			break;
	}

  return 1;
}

int GSMSMS::GetIMSI(char *imsi) {
  GetATResult("AT+CIMI",NULL,imsi,1000);
  return 1;
}

int GSMSMS::GetNetworkOperator(char *Operator) {
	
  GetATResult("AT+COPS?", "+COPS: ", Operator, 1000);
  TrimMessage(Operator, '\"', '\"');

  return 1;
}

int GSMSMS::GetNetworkMode(char *Mode) {
int imode;
	
	GetATResult("AT+CNSMOD?","+CNSMOD: 0,",Mode,1000);

	imode=atoi(Mode);
	strcpy(Mode, _NetMode[imode]);

  return 1;
}

int GSMSMS::GetServiceCenter(char *SCA) {
  	
  // Service Center Address
  GetATResult("AT+CSCA?", "+CSCA: ", SCA, 1000);
  TrimMessage(SCA, '\"', '\"');

  return 1;
}

int GSMSMS::GetPreferredMode() {
char cmode[80];
int imode;

  GetATResult("AT+CNMP?","+CNMP: ",cmode,1000);
	imode=atoi(cmode);

  return imode;
}

int GSMSMS::SetPreferredMode(int Mode) {
	char cmd[80];
	
	sprintf(cmd,"AT+CNMP=%d",Mode);
	SendCmd(cmd);

  return 1;
}

int GSMSMS::GetATResult(const char *acmd, const char *akey, char *avalue, int timeout) {
int i;
char c;
char *p1;
char *p2;

    SendCmdParams(acmd,timeout,1000,"OK",10);
    if(akey == NULL) {
      for (i=0; i<GSMBUFFSIZE; i++) {
        c=GSMCommBuff[i];
        if (c == ' ') continue;
        else if (c == '\r') continue;
        else if (c == '\n') continue;
        
        p1=GSMCommBuff+i;
        break;
      }
    } else {
      p1=strstr(GSMCommBuff,akey);
    }
    if(p1 == NULL) {
      strcpy(avalue, "");
      return 0;
    }
    if(akey != NULL) p1 += strlen(akey);
    p2=strchr(p1,'\r');
    if(p2 != NULL) *p2='\0';
    p2=strchr(p1,'\n');
    if(p2 != NULL) *p2='\0';

    strcpy(avalue,p1);

    return 1;
}

int GSMSMS::TrimMessage(char *Msg, char StartDelim, char EndDelim) {
char *p1;
char *p2;
char *p3;
	
  p1=strchr(Msg, StartDelim);
  if(p1 != NULL) {
    p1++;
    p2=strchr(p1, EndDelim);
    if(p2 != NULL) {
      *p2='\0';
      p3=p1;
    } else p3=Msg;
  } else p3=Msg;
  	
  strcpy(Msg, p3);			// Overlapping
  
  return 1;
}

// ----------------------------- GSMSMS Library GPS Functions ------------------------------------------------------------------

#ifdef GSM_GPS

int GSMSMS::EnableGPS() {
	switch (_SimCom) {
		case 900:
			return 0;
		case 808:
			SendCmd("AT+CGNSPWR=1");
			SendCmd("AT+CGNSSEQ=RMC");
			break;
		case 5320:
		case 7100:
		case 7600:
			SendCmd("AT+CGPS=1");
			break;
	}
	return 1;
}

int GSMSMS::DisableGPS() {
	switch (_SimCom) {
		case 900:
			return 0;
		case 808:
			SendCmd("AT+CGNSPWR=0");
			SendCmd("AT+CGNSSEQ=RMC");
			break;
		case 5320:
		case 7100:
		case 7600:
			SendCmd("AT+CGPS=0");
			break;
	}
	return 1;
}

int GSMSMS::GetGPSPosition(char *latitude, char *longitude) {
	switch (_SimCom) {
		case 900:
			strcpy(latitude,"N/A");
			strcpy(longitude,"N/A");
			break;
		case 808:
			return GetGPSPosition808(latitude, longitude);
		case 5320:
		case 7100:
		case 7600:
			return GetGPSPosition5320(latitude, longitude);
	}
	return 0;
}

int GSMSMS::GetGPSPosition5320(char *latitude, char *longitude) {
char *p1;
char *p2;
char *lat_raw;
char *lon_raw;
char lat[20];
char lon[20];
char latlon[20];
char latns;
char lonew;
char minutes[10];
char degrees[4];
float fix;
unsigned long GPSTimer;
unsigned long GPSTimeout=120000l;
int ConfirmedOn;

//  AT+CGPSINFO Field Info (Comma Delimited)
//																										Length
//  "+CGPSINFO: "				(Header)
// 	Latitude,						(ddmm.mmmmmm)								  11
//  Lat N/S,            (N or S)											1
//	Longitude,					(dddmm.mmmmmm)							  12
//  Long E/W,           (E or W)											1
//  Date,								(ddmmyy)											6
//  UTC Time,						(hhmmss.s)										8
//	MSL Altitude,				(meters: float)											
//	Speed Over Ground,	(knots: float)	    						
//	Course Over Ground,	(degrees: int)

//  (Next Line)
//  AmpI/AmpQ:					AmpI/AmpQ: {AmpI}/{AmpQ}

	ConfirmedOn=0;
	GPSTimer=millis();
	while (1) {
		SendCmd("AT+CGPSINFO");
		
		// Check to see if we have a fix yet  (Will return "+CGPSINFO:,,,,,,,," or "+CGPSINFO: ,,,,,,,," if no fix)
	  p1=strstr(GSMCommBuff,",,,,,,");
  	if(p1 == NULL) break;											       // It appears we have a fix!
  		
		if(!ConfirmedOn) {															// Try Turning On GPS (Just in case)
			EnableGPS();
			ConfirmedOn++;
		}
		
		delay(5000);
		
		if((GPSTimer + GPSTimeout) < millis()) return 0;
	}
	
	// Locate LATITUDE between FIRST colon and FIRST comma (ddmm.mmmmmm)
	p1=strchr(GSMCommBuff,':');
	if(p1 == NULL) return 0;
	p1++;
	if(*p1 == ' ') p1++;						// 7100 appears to have a " " immediately after the ":" (undocumented)
	p2=strchr(p1,',');
	if(p2 == NULL) return 0;
	*p2='\0';
  strcpy(lat,p1);
	
	// Locate LATITUDE N/S between FIRST comma and SECOND comma
  p1=p2+1;
	p2=strchr(p1,',');
	if(p2 == NULL) return 0;
	latns=*p1;
	
  // Locate LONGITUDE between SECOND comma and THIRD comma (dddmm.mmmmmm)
  p1=p2+1;	
  p2=strchr(p1,',');
  if(p2 == NULL) return 0;
  *p2='\0';
  strcpy(lon,p1);
  
	// Locate LONGITUDE E/W between THIRD comma and FOURTH comma
  p1=p2+1;
	p2=strchr(p1,',');
	if(p2 == NULL) return 0;
	lonew=*p1;
	
	// Convert LATITUDE ddmm.mmmmmm N/S to (+/-)dd.dddddd
	strcpy(minutes,lat+2);
	p1=lat+2;
	*p1 ='\0';
	strcpy(degrees,lat);
	fix=atof(degrees)+(atof(minutes)/60.0);
	if (fix < 10.0) dtostrf(fix,8,6,latlon);
	else dtostrf(fix,9,6,latlon);
	if(latns == 'N' ) strcpy(latitude,"+");
	else strcpy(latitude,"-");
	strcat(latitude,latlon);

	// Convert LONGITUDE dddmm.mmmmmm E/W to (+/-)ddd.dddddd 
	strcpy(minutes,lon+3);
	p1=lon+3;
	*p1 ='\0';
	strcpy(degrees,lon);
	fix=atof(degrees)+(atof(minutes)/60.0);
	if (fix < 10.0) dtostrf(fix,8,6,latlon);
	else if (fix < 100.0) dtostrf(fix,9,6,latlon);
	else dtostrf(fix,10,6,latlon);
	if(lonew == 'E' ) strcpy(longitude,"+");
	else strcpy(longitude,"-");
	strcat(longitude,latlon);

  return 1;
}

int GSMSMS::GetGPSPosition808(char *latitude, char *longitude) {
char *p1;
char *p2;
char *lat_raw;
char *lon_raw;
unsigned long GPSTimer;
unsigned long GPSTimeout=120000l;

//  AT+CGNSINF Field Info (RMC) (Comma Delimited)
//																										Length
//  "+CGNSINF: "				(Header)
// 	GNSS Run Status,		(0=Off, 1=On)									1
// 	Fix Status,					(0=None, 1=Good)							1
// 	UTC Date and Time,	(yyyyMMddhhmmss.sss)					18
// 	Latitude,						(+/-dd.dddddd)								10
//	Longitude,					(+/-ddd.dddddd)								11
//	MSL Altitude,				(meters)											8
//	Spped Over Ground,	(km/hr: 0-999.99)							6
//	Course Over Ground,	(degrees: 0-360.0)						6
//	Fix Mode,						(0,1,2)												1
//	Reserved1,																				0
//	HDOP,								(0-99.9)											4
//	PDOP,								(0-99.9)											4
//	VDOP,								(0-99.9)											4
//	Reserved2,																				0
//	GNSS Sats in View,	(0-99)												2
//	GNSS Sats Used,			(0-99)												2
//	GLONASS Sats Used,	(0-99)												2
//	Reserved3,																				0
//	C/N0 Max,						(dBHz)												2
//	HPA,								(meters)											6
//	VPA									(meters												6

	GPSTimer=millis();
	while (1) {
		SendCmd("AT+CGNSINF");
		
		// GPS Power Status is character before FIRST Comma (0 = Off, 1 = On)
	  p1=strchr(GSMCommBuff,',');
  	if(p1 == NULL) return 0;
  	p1--;
		if(*p1 == '0') {
			EnableGPS();
			delay(5000);
		}
	
		// Fix Status character before SECOND Comma (O = no fix, 1 = fix)  Wait for fix...
		p2=p1+2;
		p1=strchr(p2,',');
	  if(p1 == NULL) return 0;
		p1--;
		if(*p1 == '0') delay(5000);
		else break;																			// Good to go!  We have a fix!
			
		if((GPSTimer + GPSTimeout) < millis()) return 0;
	}

  // Locate LATITUDE between THIRD comma and FOURTH comma (+/-dd.dddddd)
  p2=p1+2;
	p1=strchr(p2,',');
	if(p1 == NULL) return 0;
	p1++;
	p2=strchr(p1,',');
	if(p2 == NULL) return 0;
	*p2='\0';
  strcpy(latitude,p1);
	
  // Locate LONGITUDE between FOURTH comma and FIFTH comma (+/-ddd.dddddd)
  p1=++p2;	
  p2=strchr(p1,',');
  if(p2 == NULL) return 0;
  *p2='\0';
  strcpy(longitude,p1);
  
  return 1;
}

#endif

// ----------------------------- GSMSMS Library Routines with Hardware Dependencies --------------------------------------------

//  ESP32 Hardware Serial ------------------------------------------------------------------------------------------------------------

#ifdef GSM_ESP32_SERIAL

void GSMSMS::Begin(long Baud_Rate)
{
	
	if(_LastBaud != 0) SerialReset();
	
	Serial1.begin(Baud_Rate, SERIAL_8N1, _RXPin, _TXPin);
	
	_LastBaud = Baud_Rate;
	
	return;  
}

void GSMSMS::SerialReset()
{
	Serial1.flush ();   		// wait for send buffer to empty
	delay (2);    					// let last character be sent
	Serial1.end ();      		// close serial
}

int GSMSMS::Available()
{
	return Serial1.available();
}

int GSMSMS::Read()
{
	return Serial1.read();
}

int GSMSMS::Write(int c)
{
	return Serial1.write(c);
}

long GSMSMS::Print(const char *text)
{
	return Serial1.print(text);
}

long GSMSMS::Print(int i)
{
	return Serial1.print(i);
}

long GSMSMS::Println(const char *text)
{
	return Serial1.println(text);
}

long GSMSMS::Println(int i)
{
	return Serial1.println(i);
}

#endif

//  Hardware Serial ------------------------------------------------------------------------------------------------------------

#ifdef GSM_HW_SERIAL

void GSMSMS::Begin(long Baud_Rate)
{
	switch (_UART) {
		case 1:
			return Serial1.begin(Baud_Rate);   
		case 2:
			return Serial2.begin(Baud_Rate);  
	}	  
}

void GSMSMS::SerialReset()
{
	switch (_UART) {
		case 1:
				Serial1.flush ();   		// wait for send buffer to empty
				delay (2);    					// let last character be sent
				Serial1.end ();      		// close serial
		case 2:
				Serial2.flush ();   		// wait for send buffer to empty
				delay (2);    					// let last character be sent
				Serial2.end ();      		// close serial
	}	  
}

int GSMSMS::Available()
{
	switch (_UART) {
		case 1:
  		return Serial1.available();
		case 2:
  		return Serial2.available();
	}	  
}

int GSMSMS::Read()
{
	switch (_UART) {
		case 1:
		  return Serial1.read();
		case 2:
		  return Serial2.read();
	}	  
}

int GSMSMS::Write(int c)
{
	switch (_UART) {
		case 1:
		  return Serial1.write(c);
		case 2:
		  return Serial2.write(c);
	}	  
}

long GSMSMS::Print(const char *text)
{
	switch (_UART) {
		case 1:
  		return Serial1.print(text);
		case 2:
  		return Serial2.print(text);
	}	  
}

long GSMSMS::Print(int i)
{
	switch (_UART) {
		case 1:
  		return Serial1.print(i);
		case 2:
  		return Serial2.print(i);
	}	  
}

long GSMSMS::Println(const char *text)
{
	switch (_UART) {
		case 1:
  		return Serial1.println(text);
		case 2:
  		return Serial2.println(text);
	}	  
}

long GSMSMS::Println(int i)
{
	switch (_UART) {
		case 1:
  		return Serial1.println(i);
		case 2:
  		return Serial2.println(i);
	}	  
}

#endif

//  Software Serial ------------------------------------------------------------------------------------------------------------

#ifdef GSM_SW_SERIAL

void GSMSMS::Begin(long Baud_Rate)
{
  return GSMSS.begin(Baud_Rate);
}

void GSMSMS::SerialReset()
{
	GSMSS.flush ();   			// wait for send buffer to empty
	delay (2);    					// let last character be sent
	GSMSS.end ();      		// close serial
}

int GSMSMS::Available()
{
  return GSMSS.available();
}

int GSMSMS::Read()
{
  return GSMSS.read();
}

int GSMSMS::Write(int c)
{
  return GSMSS.write(c);
}

long GSMSMS::Print(const char *text)
{
  return GSMSS.print(text);
}

long GSMSMS::Print(int i)
{
  return GSMSS.print(i);
}

long GSMSMS::Println(const char *text)
{
  return GSMSS.println(text);
}

long GSMSMS::Println(int i)
{
  return GSMSS.println(i);
}

#endif